/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "cpu.h"

/* thanks to wmfire */
u_int8 getCpuLoad ( )
{
	static int lastloadstuff = 0, lastidle = 0, lastnice = 0;
	int loadline = 0, loadstuff = 0, nice = 0, idle = 0;

	FILE *stat = fopen("/proc/stat", "r");
	char temp[128], *p;

	if (!stat)
	{
		static bool errorFlag = false;
		if( !errorFlag ) {
			PRINTQ( stderr, "Error reading the file '/proc/stat' for cpu-load info\n" );
			errorFlag = true;
		}
		return 100;
	}
	
	while (fgets(temp, 128, stat)) {
		if (!strncmp(temp, "cpu", 3) && temp[3] == ' ' )
		{
			p = strtok(temp, " \t\n");
			loadstuff = atol(p = strtok(NULL, " \t\n"));
			nice = atol(p = strtok(NULL, " \t\n"));
			loadstuff += atol(p = strtok(NULL, " \t\n"));
			idle = atol(p = strtok(NULL, " \t\n"));
			break;
		}
	}

	fclose(stat);

	if (!lastloadstuff && !lastidle && !lastnice) {
		lastloadstuff = loadstuff; lastidle = idle; lastnice = nice;
		return 0;
	}

	loadline = (loadstuff-lastloadstuff)+(idle-lastidle);
	
	if (loadline)
		loadline = ((loadstuff-lastloadstuff)*100)/loadline;
	else loadline = 100;
	
	lastloadstuff = loadstuff; lastidle = idle; lastnice = nice;
	
	return loadline;
}


void manageCpuLoad( )
{
	static u_int8 lastCpuLoad = 100;
	static u_int32 lastFreqChange = 0;
	
	/* update cpu value and redraw */
	cpuLoad = getCpuLoad();

	
	if( cpuLoad != lastCpuLoad || cpuLoad == 100 )
		draw_all();
	
	lastCpuLoad = cpuLoad;
	
	
	
	if(( cpuState.auto_freq_state == false ) ||
	   ( cpuLoad == 100 && cpuState.actualFreq == cpuState.maxFreq ) ||
	   ( cpuLoad != 100 && cpuState.actualFreq == cpuState.minFreq ))
		return;
#ifdef DEBUG
	printf("AutoFreq = %d, cpuLoad = %d, actualFreq = %d, maxFreq = %d, minFreq = %d\n",
	   cpuState.auto_freq_state, cpuLoad, cpuState.actualFreq, cpuState.maxFreq, cpuState.minFreq );
#endif
		
		/* if auto_freq_state is true, increase a time per second
	 * following cpu frequency by a unit until max frequency possible */
	
	/* a time per second.. */
	if( lastFreqChange == secondsCounter )
		return;
	
	/* first time will wait for a few seconds */
	if( lastFreqChange != secondsCounter-1 )
	{
		lastFreqChange = secondsCounter;
		return;
	}
	
	lastFreqChange = secondsCounter;
	
	/* increase if cpuload is 100 and if it isn't already at its maximum.. */
	if( cpuLoad == 100 && cpuState.actualFreq < cpuState.maxFreq )
		cpuSetFreq( FREQ_INCREMENT, FREQ_STEP );
	
	/* .. decrease if cpuload isn't 100 and if it isn't already at its minimum */
	if( cpuLoad != 100 && cpuState.actualFreq > cpuState.minFreq )
		cpuSetFreq( FREQ_DECREMENT, FREQ_STEP );
	
	
	/* let's read actual frequency to avoid a two seconds wait
	 * dto now it */
	cpuReadFreq();
	
	return;
}



void cpuReadFreq( )
{
	FILE * fp;
	char red[10];
	
	if( (fp = fopen( cpuState.setFreqFile, "r" )) == NULL )
	{
		fprintf(stderr, "Error reading the info file (%s):\n%s\n", cpuState.setFreqFile, strerror(errno) );
		free_and_exit( ERROR );
	}
	
	fgets ( red, 10, fp );
	cpuState.actualFreq = atoi ( red );
	
	fclose ( fp );
}



void cpuCheckFreq()
{
	static u_int32 lastCpuFreq = 0;
	
	cpuReadFreq();
	
	if( lastCpuFreq != cpuState.actualFreq )
		draw_all();
	
	lastCpuFreq = cpuState.actualFreq;
}


void cpuSetFreq( bool direction, bool speed )
{
	/* let's start with a required control */
	if(( direction == FREQ_INCREMENT && cpuState.actualFreq == cpuState.maxFreq ) ||
	   ( direction == FREQ_DECREMENT && cpuState.actualFreq == cpuState.minFreq ))
		return;
	
#ifdef DEBUG
	printf("cpuSetFreq() called: direction %u, speed %u, seconds %ld\n",
	        direction, speed, secondsCounter );
#endif
	
	switch( speed )
	{
		case FREQ_STEP:
			switch( direction )
			{
				case FREQ_INCREMENT:
					cpuState.setFreq += cpuState.stepFreq;
					if( cpuState.setFreq > cpuState.maxFreq )
						cpuState.setFreq = cpuState.maxFreq;
					break;
				case FREQ_DECREMENT:
					cpuState.setFreq -= cpuState.stepFreq;
					if( cpuState.setFreq < cpuState.minFreq )
						cpuState.setFreq = cpuState.minFreq;
			}
			
			cpuEchoFreq( );
			break;
		
			
		case FREQ_NEXT:
		{
			/* is speed is FREQ_NEXT we shall increase/decrease frequency
			 * until we find that's really changed when we go to re-read it.
			 * ( it is the behaviour that we want when we clic on frequency
			 * arrows) */
			u_int32 lastCpuFreq = cpuState.actualFreq;
			
			while( lastCpuFreq == cpuState.actualFreq )
			{
				switch( direction )
				{
					case FREQ_INCREMENT:
						cpuState.setFreq += FREQ_NEXT_DEFAULT;
						if( cpuState.setFreq > cpuState.maxFreq )
							cpuState.setFreq = cpuState.maxFreq;
						break;
					case FREQ_DECREMENT:
						cpuState.setFreq -= FREQ_NEXT_DEFAULT;
						if( cpuState.setFreq < cpuState.minFreq )
							cpuState.setFreq = cpuState.minFreq;
				}
				
				cpuEchoFreq();
				cpuReadFreq();
			}
			
			cpuState.setFreq = cpuState.actualFreq;
		}
	} /* switch( speed ) */
		
}


void cpuEchoFreq()
{
	FILE * fp;
	
#ifdef DEBUG
	printf("cpuEchoFreq() called: %u\n", cpuState.setFreq );
#endif
	
	if( (fp = fopen( cpuState.setFreqFile, "w" )) == NULL )
	{
		fprintf(stderr, "Error writing the new freq (%d) in the file (%s):\n%s\n", \
		                cpuState.setFreq, cpuState.setFreqFile, strerror(errno) );
		free_and_exit( ERROR );
	}
	
	fprintf( fp, "%u", cpuState.setFreq );
	
	fclose ( fp );
}


/* it performs two checks:
 * - passing from 99% to 100% battery state (AC is plugged in) will mean
 *   to set the max cpu freq and disabling the autofreq (this works even
 *   if we are at 100% battery percentage while wmlaptop starts)
 * - passing from 100% to 99% after the pass 1 has been performed will
 *   mean to enable the autofreq
*/
void cpuSetParadisiac( )
{
	static bool pass1done = false;
	static u_int8 lastPercentage = 101;
	static bool saved_autoFreq = false;

	if( pass1done && powerState.isCharging == false )
	{
		pass1done = false;
		cpuState.auto_freq_state = saved_autoFreq;
	}
	
	if( ( lastPercentage == 99 || lastPercentage == 101 ) && powerState.percentage == 100 && powerState.isCharging )
	{
		/* set cpu freq to max */
		while( cpuState.actualFreq < cpuState.maxFreq )
		{
			cpuSetFreq( FREQ_INCREMENT, FREQ_NEXT );
			cpuReadFreq( );
		}
		
        saved_autoFreq = cpuState.auto_freq_state;
		cpuState.auto_freq_state = false;
		pass1done = true;
	}
	
	lastPercentage = powerState.percentage;
		
}
